from coredis.exceptions import RedisError
from coredis.utils import b, nativestr


def parse_georadius_generic(response, **options):
    if options["store"] or options["store_dist"]:
        # `store` and `store_diff` cant be combined
        # with other command arguments.
        return response

    if type(response) != list:
        response_list = [response]
    else:
        response_list = response

    if not options["withdist"] and not options["withcoord"] and not options["withhash"]:
        # just a bunch of places
        return [nativestr(r) for r in response_list]

    cast = {
        "withdist": float,
        "withcoord": lambda ll: (float(ll[0]), float(ll[1])),
        "withhash": int,
    }

    # zip all output results with each casting functino to get
    # the properly native Python value.
    f = [nativestr]
    f += [cast[o] for o in ["withdist", "withhash", "withcoord"] if options[o]]
    return [list(map(lambda fv: fv[0](fv[1]), zip(f, r))) for r in response_list]


class GeoCommandMixin:

    RESPONSE_CALLBACKS = {
        "GEOPOS": lambda r: list(
            map(lambda ll: (float(ll[0]), float(ll[1])) if ll is not None else None, r)
        ),
        "GEOHASH": lambda r: list(r),
        "GEORADIUS": parse_georadius_generic,
        "GEORADIUSBYMEMBER": parse_georadius_generic,
        "GEODIST": float,
        "GEOADD": int,
    }

    # GEO COMMANDS
    async def geoadd(self, name, *values):
        """
        Add the specified geospatial items to the specified key identified
        by the ``name`` argument. The Geospatial items are given as ordered
        members of the ``values`` argument, each item or place is formed by
        the triad latitude, longitude and name.
        """
        if len(values) % 3 != 0:
            raise RedisError("GEOADD requires places with lon, lat and name" " values")
        return await self.execute_command("GEOADD", name, *values)

    async def geodist(self, name, place1, place2, unit=None):
        """
        Return the distance between ``place1`` and ``place2`` members of the
        ``name`` key.
        The units must be one of the following : m, km mi, ft. By async default
        meters are used.
        """
        pieces = [name, place1, place2]
        if unit and unit not in ("m", "km", "mi", "ft"):
            raise RedisError("GEODIST invalid unit")
        elif unit:
            pieces.append(unit)
        return await self.execute_command("GEODIST", *pieces)

    async def geohash(self, name, *values):
        """
        Return the geo hash string for each item of ``values`` members of
        the specified key identified by the ``name``argument.
        """
        return await self.execute_command("GEOHASH", name, *values)

    async def geopos(self, name, *values):
        """
        Return the positions of each item of ``values`` as members of
        the specified key identified by the ``name``argument. Each position
        is represented by the pairs lon and lat.
        """
        return await self.execute_command("GEOPOS", name, *values)

    async def georadius(
        self,
        name,
        longitude,
        latitude,
        radius,
        unit=None,
        withdist=False,
        withcoord=False,
        withhash=False,
        count=None,
        sort=None,
        store=None,
        store_dist=None,
    ):
        """
        Return the members of the specified key identified by the
        ``name`` argument which are within the borders of the area specified
        with the ``latitude`` and ``longitude`` location and the maximum
        distance from the center specified by the ``radius`` value.

        The units must be one of the following : m, km mi, ft. By default

        ``withdist`` indicates to return the distances of each place.

        ``withcoord`` indicates to return the latitude and longitude of
        each place.

        ``withhash`` indicates to return the geohash string of each place.

        ``count`` indicates to return the number of elements up to N.

        ``sort`` indicates to return the places in a sorted way, ASC for
        nearest to fairest and DESC for fairest to nearest.

        ``store`` indicates to save the places names in a sorted set named
        with a specific key, each element of the destination sorted set is
        populated with the score got from the original geo sorted set.

        ``store_dist`` indicates to save the places names in a sorted set
        named with a specific key, instead of ``store`` the sorted set
        destination score is set with the distance.
        """
        return await self._georadiusgeneric(
            "GEORADIUS",
            name,
            longitude,
            latitude,
            radius,
            unit=unit,
            withdist=withdist,
            withcoord=withcoord,
            withhash=withhash,
            count=count,
            sort=sort,
            store=store,
            store_dist=store_dist,
        )

    async def georadiusbymember(
        self,
        name,
        member,
        radius,
        unit=None,
        withdist=False,
        withcoord=False,
        withhash=False,
        count=None,
        sort=None,
        store=None,
        store_dist=None,
    ):
        """
        This command is exactly like ``georadius`` with the sole difference
        that instead of taking, as the center of the area to query, a longitude
        and latitude value, it takes the name of a member already existing
        inside the geospatial index represented by the sorted set.
        """
        return await self._georadiusgeneric(
            "GEORADIUSBYMEMBER",
            name,
            member,
            radius,
            unit=unit,
            withdist=withdist,
            withcoord=withcoord,
            withhash=withhash,
            count=count,
            sort=sort,
            store=store,
            store_dist=store_dist,
        )

    async def _georadiusgeneric(self, command, *args, **kwargs):
        pieces = list(args)
        if kwargs["unit"] and kwargs["unit"] not in ("m", "km", "mi", "ft"):
            raise RedisError("GEORADIUS invalid unit")
        elif kwargs["unit"]:
            pieces.append(kwargs["unit"])
        else:
            pieces.append("m",)

        for token in ("withdist", "withcoord", "withhash"):
            if kwargs[token]:
                pieces.append(b(token.upper()))

        if kwargs["count"]:
            pieces.extend([b("COUNT"), kwargs["count"]])

        if kwargs["sort"] and kwargs["sort"] not in ("ASC", "DESC"):
            raise RedisError("GEORADIUS invalid sort")
        elif kwargs["sort"]:
            pieces.append(b(kwargs["sort"]))

        if kwargs["store"] and kwargs["store_dist"]:
            raise RedisError("GEORADIUS store and store_dist cant be set" " together")

        if kwargs["store"]:
            pieces.extend([b("STORE"), kwargs["store"]])

        if kwargs["store_dist"]:
            pieces.extend([b("STOREDIST"), kwargs["store_dist"]])

        return await self.execute_command(command, *pieces, **kwargs)



#!/usr/bin/python
# -*- coding: utf-8 -*-
from collections import defaultdict


class IterCommandMixin:
    """
    convenient function of scan iter, make it a class separately
    because yield can not be used in async function in Python3.6
    """

    RESPONSE_CALLBACKS = {}

    async def scan_iter(self, match=None, count=None):
        """
        Make an iterator using the SCAN command so that the client doesn't
        need to remember the cursor position.

        ``match`` allows for filtering the keys by pattern

        ``count`` allows for hint the minimum number of returns
        """
        cursor = "0"
        while cursor != 0:
            cursor, data = await self.scan(cursor=cursor, match=match, count=count)
            for item in data:
                yield item

    async def sscan_iter(self, name, match=None, count=None):
        """
        Make an iterator using the SSCAN command so that the client doesn't
        need to remember the cursor position.

        ``match`` allows for filtering the keys by pattern

        ``count`` allows for hint the minimum number of returns
        """
        cursor = "0"
        while cursor != 0:
            cursor, data = await self.sscan(
                name, cursor=cursor, match=match, count=count
            )
            for item in data:
                yield item

    async def hscan_iter(self, name, match=None, count=None):
        """
        Make an iterator using the HSCAN command so that the client doesn't
        need to remember the cursor position.

        ``match`` allows for filtering the keys by pattern

        ``count`` allows for hint the minimum number of returns
        """
        cursor = "0"
        while cursor != 0:
            cursor, data = await self.hscan(
                name, cursor=cursor, match=match, count=count
            )
            for item in data.items():
                yield item

    async def zscan_iter(self, name, match=None, count=None, score_cast_func=float):
        """
        Make an iterator using the ZSCAN command so that the client doesn't
        need to remember the cursor position.

        ``match`` allows for filtering the keys by pattern

        ``count`` allows for hint the minimum number of returns

        ``score_cast_func`` a callable used to cast the score return value
        """
        cursor = "0"
        while cursor != 0:
            cursor, data = await self.zscan(
                name,
                cursor=cursor,
                match=match,
                count=count,
                score_cast_func=score_cast_func,
            )
            for item in data:
                yield item


class ClusterIterCommandMixin(IterCommandMixin):
    async def scan_iter(self, match=None, count=None):
        nodes = await self.cluster_nodes()
        for node in nodes:
            if "master" in node["flags"]:
                cursor = "0"
                while cursor != 0:
                    pieces = [cursor]
                    if match is not None:
                        pieces.extend(["MATCH", match])
                    if count is not None:
                        pieces.extend(["COUNT", count])
                    response = await self.execute_command_on_nodes(
                        [node], "SCAN", *pieces
                    )
                    cursor, data = list(response.values())[0]
                    for item in data:
                        yield item



# -*- coding: utf-8 -*-

# python std lib
from __future__ import with_statement
import asyncio
import uuid

# rediscluster imports
from coredis import (
    StrictRedis,
    StrictRedisCluster,
    RedisClusterException,
    ConnectionError,
)
from coredis.nodemanager import NodeManager

# 3rd party imports
import pytest
from mock import patch, Mock


def test_set_node_name(s):
    """
    Test that method sets ["name"] correctly
    """
    n = {"host": "127.0.0.1", "port": 7000}
    s.connection_pool.nodes.set_node_name(n)
    assert "name" in n
    assert n["name"] == "127.0.0.1:7000"


def test_keyslot():
    """
    Test that method will compute correct key in all supported cases
    """
    n = NodeManager([{}])

    assert n.keyslot("foo") == 12182
    assert n.keyslot("{foo}bar") == 12182
    assert n.keyslot("{foo}") == 12182
    assert n.keyslot(1337) == 4314

    assert n.keyslot(125) == n.keyslot(b"125")
    assert n.keyslot(125) == n.keyslot("\x31\x32\x35")
    assert n.keyslot("大奖") == n.keyslot(b"\xe5\xa4\xa7\xe5\xa5\x96")
    assert n.keyslot(u"大奖") == n.keyslot(b"\xe5\xa4\xa7\xe5\xa5\x96")
    assert n.keyslot(1337.1234) == n.keyslot("1337.1234")
    assert n.keyslot(1337) == n.keyslot("1337")
    assert n.keyslot(b"abc") == n.keyslot("abc")
    assert n.keyslot("abc") == n.keyslot(str("abc"))
    assert n.keyslot(str("abc")) == n.keyslot(b"abc")


@pytest.mark.asyncio
async def test_init_slots_cache_not_all_slots(s):
    """
    Test that if not all slots are covered it should raise an exception
    """
    # Create wrapper function so we can inject custom 'CLUSTER SLOTS' command result
    def get_redis_link_wrapper(*args, **kwargs):
        link = StrictRedis(host="127.0.0.1", port=7000)

        orig_exec_method = link.execute_command

        async def patch_execute_command(*args, **kwargs):
            if args == ("CLUSTER SLOTS",):
                # Missing slot 5460
                return {
                    (0, 5459): [
                        {
                            "host": "127.0.0.1",
                            "port": 7000,
                            "node_id": str(uuid.uuid4()),
                            "server_type": "master",
                        },
                        {
                            "host": "127.0.0.1",
                            "port": 7003,
                            "node_id": str(uuid.uuid4()),
                            "server_type": "slave",
                        },
                    ],
                    (5461, 10922): [
                        {
                            "host": "127.0.0.1",
                            "port": 7001,
                            "node_id": str(uuid.uuid4()),
                            "server_type": "master",
                        },
                        {
                            "host": "127.0.0.1",
                            "port": 7004,
                            "node_id": str(uuid.uuid4()),
                            "server_type": "slave",
                        },
                    ],
                    (10923, 16383): [
                        {
                            "host": "127.0.0.1",
                            "port": 7002,
                            "node_id": str(uuid.uuid4()),
                            "server_type": "master",
                        },
                        {
                            "host": "127.0.0.1",
                            "port": 7005,
                            "node_id": str(uuid.uuid4()),
                            "server_type": "slave",
                        },
                    ],
                }

            return await orig_exec_method(*args, **kwargs)

        # Missing slot 5460
        link.execute_command = patch_execute_command

        return link

    s.connection_pool.nodes.get_redis_link = get_redis_link_wrapper

    with pytest.raises(RedisClusterException) as ex:
        await s.connection_pool.initialize()

    assert str(ex.value).startswith(
        "Not all slots are covered after query all startup_nodes."
    )


@pytest.mark.asyncio
async def test_init_slots_cache_not_all_slots_not_require_full_coverage(s):
    """
    Test that if not all slots are covered it should raise an exception
    """
    # Create wrapper function so we can inject custom 'CLUSTER SLOTS' command result
    def get_redis_link_wrapper(*args, **kwargs):
        link = StrictRedis(host="127.0.0.1", port=7000, decode_responses=True)

        orig_exec_method = link.execute_command

        async def patch_execute_command(*args, **kwargs):
            if args == ("CLUSTER SLOTS",):
                # Missing slot 5460
                return {
                    (0, 5459): [
                        {
                            "host": "127.0.0.1",
                            "port": 7000,
                            "node_id": str(uuid.uuid4()),
                            "server_type": "master",
                        },
                        {
                            "host": "127.0.0.1",
                            "port": 7003,
                            "node_id": str(uuid.uuid4()),
                            "server_type": "slave",
                        },
                    ],
                    (5461, 10922): [
                        {
                            "host": "127.0.0.1",
                            "port": 7001,
                            "node_id": str(uuid.uuid4()),
                            "server_type": "master",
                        },
                        {
                            "host": "127.0.0.1",
                            "port": 7004,
                            "node_id": str(uuid.uuid4()),
                            "server_type": "slave",
                        },
                    ],
                    (10923, 16383): [
                        {
                            "host": "127.0.0.1",
                            "port": 7002,
                            "node_id": str(uuid.uuid4()),
                            "server_type": "master",
                        },
                        {
                            "host": "127.0.0.1",
                            "port": 7005,
                            "node_id": str(uuid.uuid4()),
                            "server_type": "slave",
                        },
                    ],
                }
            elif args == ("CONFIG GET", "cluster-require-full-coverage"):
                return {"cluster-require-full-coverage": "no"}
            else:
                return await orig_exec_method(*args, **kwargs)

        # Missing slot 5460
        link.execute_command = patch_execute_command

        return link

    s.connection_pool.nodes.get_redis_link = get_redis_link_wrapper

    await s.connection_pool.nodes.initialize()

    assert 5460 not in s.connection_pool.nodes.slots


@pytest.mark.asyncio
async def test_init_slots_cache(s):
    """
    Test that slots cache can in initialized and all slots are covered
    """
    good_slots_resp = {
        (0, 5460): [
            {
                "host": "127.0.0.1",
                "port": 7000,
                "node_id": str(uuid.uuid4()),
                "server_type": "master",
            },
            {
                "host": "127.0.0.1",
                "port": 7003,
                "node_id": str(uuid.uuid4()),
                "server_type": "slave",
            },
        ],
        (5461, 10922): [
            {
                "host": "127.0.0.1",
                "port": 7001,
                "node_id": str(uuid.uuid4()),
                "server_type": "master",
            },
            {
                "host": "127.0.0.1",
                "port": 7004,
                "node_id": str(uuid.uuid4()),
                "server_type": "slave",
            },
        ],
        (10923, 16383): [
            {
                "host": "127.0.0.1",
                "port": 7002,
                "node_id": str(uuid.uuid4()),
                "server_type": "master",
            },
            {
                "host": "127.0.0.1",
                "port": 7005,
                "node_id": str(uuid.uuid4()),
                "server_type": "slave",
            },
        ],
    }

    with patch.object(StrictRedis, "execute_command") as execute_command_mock:

        async def patch_execute_command(*args, **kwargs):
            if args == ("CONFIG GET", "cluster-require-full-coverage"):
                return {"cluster-require-full-coverage": "yes"}
            else:
                return good_slots_resp

        execute_command_mock.side_effect = patch_execute_command

        await s.connection_pool.nodes.initialize()
        assert len(s.connection_pool.nodes.slots) == NodeManager.RedisClusterHashSlots
        for slot_info, node_info in good_slots_resp.items():
            all_hosts = ["127.0.0.1", "127.0.0.2"]
            all_ports = [7000, 7001, 7002, 7003, 7004, 7005]
            slot_start = slot_info[0]
            slot_end = slot_info[1]
            for i in range(slot_start, slot_end + 1):
                assert len(s.connection_pool.nodes.slots[i]) == len(node_info)
                assert s.connection_pool.nodes.slots[i][0]["host"] in all_hosts
                assert s.connection_pool.nodes.slots[i][1]["host"] in all_hosts
                assert s.connection_pool.nodes.slots[i][0]["port"] in all_ports
                assert s.connection_pool.nodes.slots[i][1]["port"] in all_ports

    assert len(s.connection_pool.nodes.nodes) == 6


def test_empty_startup_nodes():
    """
    It should not be possible to create a node manager with no nodes specefied
    """
    with pytest.raises(RedisClusterException):
        NodeManager()

    with pytest.raises(RedisClusterException):
        NodeManager([])


def test_wrong_startup_nodes_type():
    """
    If something other then a list type itteratable is provided it should fail
    """
    with pytest.raises(TypeError):
        NodeManager({})


@pytest.mark.asyncio
async def test_init_slots_cache_slots_collision():
    """
    Test that if 2 nodes do not agree on the same slots setup it should raise an error.
    In this test both nodes will say that the first slots block should be bound to different
     servers.
    """

    n = NodeManager(
        startup_nodes=[
            {"host": "127.0.0.1", "port": 7000},
            {"host": "127.0.0.1", "port": 7001},
        ]
    )

    def monkey_link(host=None, port=None, *args, **kwargs):
        """
        Helper function to return custom slots cache data from different redis nodes
        """
        if port == 7000:
            result = {
                (0, 5460): [
                    {
                        "host": "127.0.0.1",
                        "port": 7000,
                        "node_id": str(uuid.uuid4()),
                        "server_type": "master",
                    },
                    {
                        "host": "127.0.0.1",
                        "port": 7003,
                        "node_id": str(uuid.uuid4()),
                        "server_type": "slave",
                    },
                ],
                (5461, 10922): [
                    {
                        "host": "127.0.0.1",
                        "port": 7001,
                        "node_id": str(uuid.uuid4()),
                        "server_type": "master",
                    },
                    {
                        "host": "127.0.0.1",
                        "port": 7004,
                        "node_id": str(uuid.uuid4()),
                        "server_type": "slave",
                    },
                ],
            }
        elif port == 7001:
            result = {
                (0, 5460): [
                    {
                        "host": "127.0.0.1",
                        "port": 7001,
                        "node_id": str(uuid.uuid4()),
                        "server_type": "master",
                    },
                    {
                        "host": "127.0.0.1",
                        "port": 7003,
                        "node_id": str(uuid.uuid4()),
                        "server_type": "slave",
                    },
                ],
                (5461, 10922): [
                    {
                        "host": "127.0.0.1",
                        "port": 7000,
                        "node_id": str(uuid.uuid4()),
                        "server_type": "master",
                    },
                    {
                        "host": "127.0.0.1",
                        "port": 7004,
                        "node_id": str(uuid.uuid4()),
                        "server_type": "slave",
                    },
                ],
            }
        else:
            result = dict()

        r = StrictRedisCluster(host=host, port=port, decode_responses=True)
        orig_execute_command = r.execute_command

        async def execute_command(*args, **kwargs):
            if args == ("CLUSTER SLOTS",):
                return result
            elif args == ("CONFIG GET", "cluster-require-full-coverage"):
                return {"cluster-require-full-coverage": "yes"}
            else:
                return orig_execute_command(*args, **kwargs)

        r.execute_command = execute_command
        return r

    n.get_redis_link = monkey_link
    with pytest.raises(RedisClusterException) as ex:
        await n.initialize()
    assert str(ex.value).startswith(
        "startup_nodes could not agree on a valid slots cache."
    ), str(ex.value)


@pytest.mark.asyncio
async def test_all_nodes():
    """
    Set a list of nodes and it should be possible to itterate over all
    """
    n = NodeManager(startup_nodes=[{"host": "127.0.0.1", "port": 7000}])
    await n.initialize()

    nodes = [node for node in n.nodes.values()]

    for i, node in enumerate(n.all_nodes()):
        assert node in nodes


@pytest.mark.asyncio
async def test_all_nodes_masters():
    """
    Set a list of nodes with random masters/slaves config and it shold be possible
    to itterate over all of them.
    """
    n = NodeManager(
        startup_nodes=[
            {"host": "127.0.0.1", "port": 7000},
            {"host": "127.0.0.1", "port": 7001},
        ]
    )
    await n.initialize()

    nodes = [node for node in n.nodes.values() if node["server_type"] == "master"]

    for node in n.all_masters():
        assert node in nodes


def test_random_startup_node():
    """
    Hard to test reliable for a random
    """
    s = [{"1": 1}, {"2": 2}, {"3": 3}]
    n = NodeManager(startup_nodes=s)
    random_node = n.random_startup_node()

    for i in range(0, 5):
        assert random_node in s


@pytest.mark.asyncio
async def test_cluster_slots_error():
    """
    Check that exception is raised if initialize can't execute
    'CLUSTER SLOTS' command.
    """
    with patch.object(StrictRedisCluster, "execute_command") as execute_command_mock:
        execute_command_mock.side_effect = Exception("foobar")

        n = NodeManager(startup_nodes=[{}])

        with pytest.raises(RedisClusterException):
            await n.initialize()


def test_set_node():
    """
    Test to update data in a slot.
    """
    expected = {
        "host": "127.0.0.1",
        "name": "127.0.0.1:7000",
        "port": 7000,
        "server_type": "master",
    }

    n = NodeManager(startup_nodes=[{}])
    assert len(n.slots) == 0, "no slots should exist"
    res = n.set_node(host="127.0.0.1", port=7000, server_type="master")
    assert res == expected
    assert n.nodes == {expected["name"]: expected}


@pytest.mark.asyncio
async def test_reset():
    """
    Test that reset method resets variables back to correct default values.
    """

    class AsyncMock(Mock):
        def __await__(self):
            future = asyncio.Future(loop=asyncio.get_event_loop())
            future.set_result(self)
            result = yield from future
            return result

    n = NodeManager(startup_nodes=[{}])
    n.initialize = AsyncMock()
    await n.reset()
    assert n.initialize.call_count == 1


@pytest.mark.asyncio
async def test_cluster_one_instance():
    """
    If the cluster exists of only 1 node then there is some hacks that must
    be validated they work.
    """
    with patch.object(StrictRedis, "execute_command") as mock_execute_command:

        async def patch_execute_command(*args, **kwargs):
            if args == ("CONFIG GET", "cluster-require-full-coverage"):
                return {"cluster-require-full-coverage": "yes"}
            else:
                return {
                    (0, 16383): [
                        {
                            "host": "",
                            "port": 7006,
                            "node_id": str(uuid.uuid4()),
                            "server_type": "master",
                        }
                    ],
                }

        mock_execute_command.side_effect = patch_execute_command

        n = NodeManager(startup_nodes=[{"host": "127.0.0.1", "port": 7006}])
        await n.initialize()

        del n.nodes["127.0.0.1:7006"]["node_id"]
        assert n.nodes == {
            "127.0.0.1:7006": {
                "host": "127.0.0.1",
                "name": "127.0.0.1:7006",
                "port": 7006,
                "server_type": "master",
            }
        }

        assert len(n.slots) == 16384
        for i in range(0, 16384):
            assert n.slots[i] == [
                {
                    "host": "127.0.0.1",
                    "name": "127.0.0.1:7006",
                    "port": 7006,
                    "server_type": "master",
                }
            ]


@pytest.mark.asyncio
async def test_initialize_follow_cluster():
    n = NodeManager(
        nodemanager_follow_cluster=True,
        startup_nodes=[{"host": "127.0.0.1", "port": 7000}],
    )
    n.orig_startup_nodes = None
    await n.initialize()


@pytest.mark.asyncio
async def test_init_with_down_node():
    """
    If I can't connect to one of the nodes, everything should still work.
    But if I can't connect to any of the nodes, exception should be thrown.
    """

    def get_redis_link(host, port, decode_responses=False):
        if port == 7000:
            raise ConnectionError("mock connection error for 7000")
        return StrictRedis(host=host, port=port, decode_responses=decode_responses)

    with patch.object(NodeManager, "get_redis_link", side_effect=get_redis_link):
        n = NodeManager(startup_nodes=[{"host": "127.0.0.1", "port": 7000}])
        with pytest.raises(RedisClusterException) as e:
            await n.initialize()
        assert "Redis Cluster cannot be connected" in str(e.value)



import asyncio
import warnings
from coredis.exceptions import RedisClusterException, WatchError
from coredis.utils import string_keys_to_dict, bool_ok

from coredis.utils import LOOP_DEPRECATED


class TransactionCommandMixin:

    RESPONSE_CALLBACKS = string_keys_to_dict("WATCH UNWATCH", bool_ok)

    async def transaction(self, func, *watches, **kwargs):
        """
        Convenience method for executing the callable `func` as a transaction
        while watching all keys specified in `watches`. The 'func' callable
        should expect a single argument which is a Pipeline object.
        """
        shard_hint = kwargs.pop("shard_hint", None)
        value_from_callable = kwargs.pop("value_from_callable", False)
        watch_delay = kwargs.pop("watch_delay", None)
        async with await self.pipeline(True, shard_hint) as pipe:
            while True:
                try:
                    if watches:
                        await pipe.watch(*watches)
                    func_value = await func(pipe)
                    exec_value = await pipe.execute()
                    return func_value if value_from_callable else exec_value
                except WatchError:
                    if watch_delay is not None and watch_delay > 0:
                        if LOOP_DEPRECATED:
                            await asyncio.sleep(watch_delay)
                        else:
                            await asyncio.sleep(
                                watch_delay, loop=self.connection_pool.loop
                            )
                    continue

    async def watch(self, *names):
        """
        Watches the values at keys ``names``, or None if the key doesn't exist
        """
        warnings.warn(DeprecationWarning("Call WATCH from a Pipeline object"))

    async def unwatch(self):
        """
        Unwatches the value at key ``name``, or None of the key doesn't exist
        """
        warnings.warn(DeprecationWarning("Call UNWATCH from a Pipeline object"))


class ClusterTransactionCommandMixin(TransactionCommandMixin):
    async def transaction(self, func, *watches, **kwargs):
        """
        Convenience method for executing the callable `func` as a transaction
        while watching all keys specified in `watches`. The 'func' callable
        should expect a single argument which is a Pipeline object.

        cluster transaction can only be run with commands in the same node,
        otherwise error will be raised.
        """
        shard_hint = kwargs.pop("shard_hint", None)
        value_from_callable = kwargs.pop("value_from_callable", False)
        watch_delay = kwargs.pop("watch_delay", None)
        async with await self.pipeline(True, shard_hint, watches=watches) as pipe:
            while True:
                try:
                    func_value = await func(pipe)
                    exec_value = await pipe.execute()
                    return func_value if value_from_callable else exec_value
                except WatchError:
                    if watch_delay is not None and watch_delay > 0:
                        if LOOP_DEPRECATED:
                            await asyncio.sleep(watch_delay)
                        else:
                            await asyncio.sleep(
                                watch_delay, loop=self.connection_pool.loop
                            )
                    continue



from coredis.exceptions import RedisError
from coredis.utils import dict_merge, pairs_to_dict, string_keys_to_dict, bool_ok


def stream_list(response):
    result = []
    if response:
        for r in response:
            kv_pairs = r[1]
            kv_dict = dict()
            while kv_pairs and len(kv_pairs) > 1:
                kv_dict[kv_pairs.pop()] = kv_pairs.pop()
            result.append((r[0], kv_dict))
    return result


def multi_stream_list(response):
    result = dict()
    if response:
        for r in response:
            result[r[0]] = stream_list(r[1])
    return result


def list_of_pairs_to_dict(response):
    return [pairs_to_dict(row) for row in response]


def parse_xinfo_stream(response):
    res = pairs_to_dict(response)
    if res["first-entry"] and len(res["first-entry"]) > 0:
        res["first-entry"][1] = pairs_to_dict(res["first-entry"][1])
    if res["last-entry"] and len(res["last-entry"]) > 0:
        res["last-entry"][1] = pairs_to_dict(res["last-entry"][1])
    return res


class StreamsCommandMixin:
    RESPONSE_CALLBACKS = dict_merge(
        string_keys_to_dict("XREVRANGE XRANGE", stream_list),
        string_keys_to_dict("XREAD XREADGROUP", multi_stream_list),
        {
            "XINFO GROUPS": list_of_pairs_to_dict,
            "XINFO STREAM": parse_xinfo_stream,
            "XINFO CONSUMERS": list_of_pairs_to_dict,
            "XGROUP SETID": bool_ok,
            "XGROUP CREATE": bool_ok,
        },
    )

    async def xadd(
        self, name: str, entry: dict, max_len=None, stream_id="*", approximate=True
    ) -> str:
        """
        Appends the specified stream entry to the stream at the specified key.
        If the key does not exist, as a side effect of running
        this command the key is created with a stream value.
        Available since 5.0.0.
        Time complexity: O(log(N)) with N being the number of items already into the stream.

        :param name: name of the stream
        :param entry: key-values to be appended to the stream
        :param max_len: max length of the stream
        length will not be limited max_len is set to None
        notice: max_len should be int greater than 0,
        if set to 0 or negative, the stream length will not be limited

        :param stream_id: id of the options appended to the stream.
        The XADD command will auto-generate a unique id for you
        if the id argument specified is the * character.
        ID are specified by two numbers separated by a "-" character

        :param approximate: whether redis will limit
        the stream with given max length exactly, if set to True,
        there will be a few tens of entries more,
        but never less than 1000 items

        :return: id auto generated or the specified id given.
        notice: specified id without "-" character will be completed like "id-0"
        """
        pieces = []
        if max_len is not None:
            if not isinstance(max_len, int) or max_len < 1:
                raise RedisError("XADD maxlen must be a positive integer")
            pieces.append("MAXLEN")
            if approximate:
                pieces.append("~")
            pieces.append(str(max_len))
        pieces.append(stream_id)
        for kv in entry.items():
            pieces.extend(list(kv))
        return await self.execute_command("XADD", name, *pieces)

    async def xlen(self, name: str) -> int:
        """
        Returns the number of elements in a given stream.
        """
        return await self.execute_command("XLEN", name)

    async def xrange(self, name: str, start="-", end="+", count=None) -> list:
        """
        Read stream values within an interval.

        Available since 5.0.0.
        Time complexity: O(log(N)+M) with N being the number of elements in the stream and M the number
        of elements being returned. If M is constant (e.g. always asking for the first 10 elements with COUNT),
        you can consider it O(log(N)).

        :param name: name of the stream.
        :param start: first stream ID. defaults to '-',
               meaning the earliest available.
        :param end: last stream ID. defaults to '+',
                meaning the latest available.
        :param count: if set, only return this many items, beginning with the
               earliest available.
        :return list of (stream_id, entry(k-v pair))
        """

        pieces = [start, end]
        if count is not None:
            if not isinstance(count, int) or count < 1:
                raise RedisError("XRANGE count must be a positive integer")
            pieces.append("COUNT")
            pieces.append(str(count))
        return await self.execute_command("XRANGE", name, *pieces)

    async def xrevrange(self, name: str, start="+", end="-", count=None) -> list:
        """
        Read stream values within an interval, in reverse order.

        Available since 5.0.0.
        Time complexity: O(log(N)+M) with N being the number of elements in the stream and M the number
        of elements being returned. If M is constant (e.g. always asking for the first 10 elements with COUNT),
        you can consider it O(log(N)).

        :param name: name of the stream
        :param start: first stream ID. defaults to '+',
               meaning the latest available.
        :param end: last stream ID. defaults to '-',
                meaning the earliest available.
        :param count: if set, only return this many items, beginning with the
               latest available.

        """
        pieces = [start, end]
        if count is not None:
            if not isinstance(count, int) or count < 1:
                raise RedisError("XREVRANGE count must be a positive integer")
            pieces.append("COUNT")
            pieces.append(str(count))
        return await self.execute_command("XREVRANGE", name, *pieces)

    async def xread(self, count=None, block=None, **streams) -> dict:
        """
        Available since 5.0.0.

        Time complexity:
        For each stream mentioned: O(log(N)+M) with N being the number
        of elements in the stream and M the number of elements being returned.
        If M is constant (e.g. always asking for the first 10 elements with COUNT),
        you can consider it O(log(N)). On the other side, XADD will pay the O(N)
        time in order to serve the N clients blocked on the stream getting new data.

        Read data from one or multiple streams,
        only returning entries with an ID greater
        than the last received ID reported by the caller.

        :param count: int, if set, only return this many items, beginning with the
               earliest available.
        :param block: int, milliseconds we want to block before timing out,
                if the BLOCK option is not used, the command is synchronous
        :param streams: stream_name - stream_id mapping
        :return dict like {stream_name: [(stream_id: entry), ...]}
        """
        pieces = []
        if block is not None:
            if not isinstance(block, int) or block < 0:
                raise RedisError("XREAD block must be a positive integer")
            pieces.append("BLOCK")
            pieces.append(str(block))
        if count is not None:
            if not isinstance(count, int) or count < 1:
                raise RedisError("XREAD count must be a positive integer")
            pieces.append("COUNT")
            pieces.append(str(count))
        pieces.append("STREAMS")
        ids = []
        for partial_stream in streams.items():
            pieces.append(partial_stream[0])
            ids.append(partial_stream[1])
        pieces.extend(ids)
        return await self.execute_command("XREAD", *pieces)

    async def xreadgroup(
        self, group: str, consumer_id: str, count=None, block=None, **streams
    ):
        """
        Available since 5.0.0.

        Time complexity:
        For each stream mentioned: O(log(N)+M) with N being the number of elements
        in the stream and M the number of elements being returned.
        If M is constant (e.g. always asking for the first 10 elements with COUNT),
        you can consider it O(log(N)). On the other side,
        XADD will pay the O(N) time in order to serve
        the N clients blocked on the stream getting new data.

        Read data from one or multiple streams via the consumer group,
        only returning entries with an ID greater
        than the last received ID reported by the caller.

        :param group: the name of the consumer group
        :param consumer_id: the name of the consumer that is attempting to read
        :param count: int, if set, only return this many items, beginning with the
               earliest available.
        :param block: int, milliseconds we want to block before timing out,
                if the BLOCK option is not used, the command is synchronous
        :param streams: stream_name - stream_id mapping
        :return dict like {stream_name: [(stream_id: entry), ...]}
        """
        pieces = ["GROUP", group, consumer_id]
        if block is not None:
            if not isinstance(block, int) or block < 1:
                raise RedisError("XREAD block must be a positive integer")
            pieces.append("BLOCK")
            pieces.append(str(block))
        if count is not None:
            if not isinstance(count, int) or count < 1:
                raise RedisError("XREAD count must be a positive integer")
            pieces.append("COUNT")
            pieces.append(str(count))
        pieces.append("STREAMS")
        ids = []
        for partial_stream in streams.items():
            pieces.append(partial_stream[0])
            ids.append(partial_stream[1])
        pieces.extend(ids)
        return await self.execute_command("XREADGROUP", *pieces)

    async def xpending(
        self, name: str, group: str, start="-", end="+", count=None, consumer=None
    ) -> list:
        """
        Available since 5.0.0.

        Time complexity:
        O(log(N)+M) with N being the number of elements in the consumer
        group pending entries list, and M the number of elements being returned.
        When the command returns just the summary it runs in O(1)
        time assuming the list of consumers is small,
        otherwise there is additional O(N) time needed to iterate every consumer.

        Fetching data from a stream via a consumer group,
        and not acknowledging such data,
        has the effect of creating pending entries.
        The XPENDING command is the interface to inspect the list of pending messages.

        :param name: name of the stream
        :param group: name of the consumer group
        :param start: first stream ID. defaults to '-',
               meaning the earliest available.
        :param end: last stream ID. defaults to '+',
                meaning the latest available.
        :param count: int, number of entries
                [NOTICE] only when count is set to int,
                start & end options will have effect
                and detail of pending entries will be returned
        :param consumer: str, consumer of the stream in the group
                [NOTICE] only when count is set to int,
                this option can be appended to
                query pending entries of given consumer
        """
        pieces = [name, group]
        if count is not None:
            pieces.extend([start, end, count])
            if consumer is not None:
                pieces.append(str(consumer))
        # todo: may there be a parse function
        return await self.execute_command("XPENDING", *pieces)

    async def xtrim(self, name: str, max_len: int, approximate=True) -> int:
        """
        [NOTICE] Not officially released yet

        XTRIM is designed to accept different trimming strategies,
        even if currently only MAXLEN is implemented.

        :param name: name of the stream
        :param max_len: max length of the stream after being trimmed
        :param approximate: whether redis will limit
        the stream with given max length exactly, if set to True,
        there will be a few tens of entries more,
        but never less than 1000 items:

        :return: number of entries trimmed
        """
        pieces = ["MAXLEN"]
        if approximate:
            pieces.append("~")
        pieces.append(max_len)
        return await self.execute_command("XTRIM", name, *pieces)

    async def xdel(self, name: str, stream_id: str) -> int:
        """
        [NOTICE] Not officially released yet
        [NOTICE] In the current implementation, memory is not
        really reclaimed until a macro node is completely empty,
        so you should not abuse this feature.

        remove items from the middle of a stream, just by ID.

        :param name: name of the stream
        :param stream_id: id of the options appended to the stream.
        """
        return await self.execute_command("XDEL", name, stream_id)

    async def xinfo_consumers(self, name: str, group: str) -> list:
        """
        [NOTICE] Not officially released yet

        XINFO command is an observability interface that can be used
        with sub-commands in order to get information
        about streams or consumer groups.

        :param name: name of the stream
        :param group: name of the consumer group
        """
        return await self.execute_command("XINFO CONSUMERS", name, group)

    async def xinfo_groups(self, name: str) -> list:
        """
        [NOTICE] Not officially released yet

        XINFO command is an observability interface that can be used
        with sub-commands in order to get information
        about streams or consumer groups.

        :param name: name of the stream
        """
        return await self.execute_command("XINFO GROUPS", name)

    async def xinfo_stream(self, name: str) -> dict:
        """
        [NOTICE] Not officially released yet

        XINFO command is an observability interface that can be used
        with sub-commands in order to get information
        about streams or consumer groups.

        :param name: name of the stream
        """
        return await self.execute_command("XINFO STREAM", name)

    async def xack(self, name: str, group: str, stream_id: str) -> int:
        """
        [NOTICE] Not officially released yet

        XACK is the command that allows a consumer to mark a pending message as correctly processed.

        :param name: name of the stream
        :param group: name of the consumer group
        :param stream_id: id of the entry the consumer wants to mark
        :return: number of entry marked
        """
        return await self.execute_command("XACK", name, group, stream_id)

    async def xclaim(
        self, name: str, group: str, consumer: str, min_idle_time: int, *stream_ids
    ):
        """
        [NOTICE] Not officially released yet

        Gets ownership of one or multiple messages in the Pending Entries List of a given stream consumer group.

        :param name: name of the stream
        :param group: name of the consumer group
        :param consumer: name of the consumer
        :param min_idle_time: ms
            If the message ID (among the specified ones) exists, and its idle time greater
            or equal to min_idle_time, then the message new owner
            becomes the specified <consumer>. If the minimum idle time specified
            is zero, messages are claimed regardless of their idle time.
        :param stream_ids:
        """
        return await self.execute_command(
            "XCLAIM", name, group, consumer, min_idle_time, *stream_ids
        )

    async def xgroup_create(self, name: str, group: str, stream_id="$") -> bool:
        """
        [NOTICE] Not officially released yet
        XGROUP is used in order to create, destroy and manage consumer groups.
        :param name: name of the stream
        :param group: name of the consumer group
        :param stream_id:
            If we provide $ as we did, then only new messages arriving
            in the stream from now on will be provided to the consumers in the group.
            If we specify 0 instead the consumer group will consume all the messages
            in the stream history to start with.
            Of course, you can specify any other valid ID
        """
        return await self.execute_command("XGROUP CREATE", name, group, stream_id)

    async def xgroup_set_id(self, name: str, group: str, stream_id: str) -> bool:
        """
        [NOTICE] Not officially released yet
        :param name: name of the stream
        :param group: name of the consumer group
        :param stream_id:
            If we provide $ as we did, then only new messages arriving
            in the stream from now on will be provided to the consumers in the group.
            If we specify 0 instead the consumer group will consume all the messages
            in the stream history to start with.
            Of course, you can specify any other valid ID
        """
        return await self.execute_command("XGROUP SETID", name, group, stream_id)

    async def xgroup_destroy(self, name: str, group: str) -> int:
        """
        [NOTICE] Not officially released yet
        XGROUP is used in order to create, destroy and manage consumer groups.
        :param name: name of the stream
        :param group: name of the consumer group
        """
        return await self.execute_command("XGROUP DESTROY", name, group)

    async def xgroup_del_consumer(self, name: str, group: str, consumer: str) -> int:
        """
        [NOTICE] Not officially released yet
        XGROUP is used in order to create, destroy and manage consumer groups.
        :param name: name of the stream
        :param group: name of the consumer group
        :param consumer: name of the consumer
        """
        return await self.execute_command("XGROUP DELCONSUMER", name, group, consumer)



import os
import random
import weakref

from coredis import StrictRedis
from coredis.connection import Connection
from coredis.pool import ConnectionPool
from coredis.exceptions import (
    ConnectionError,
    ResponseError,
    ReadOnlyError,
    TimeoutError,
)
from coredis.utils import iteritems, nativestr


class MasterNotFoundError(ConnectionError):
    pass


class SlaveNotFoundError(ConnectionError):
    pass


class SentinelManagedConnection(Connection):
    def __init__(self, **kwargs):
        self.connection_pool = kwargs.pop("connection_pool")
        super(SentinelManagedConnection, self).__init__(**kwargs)

    def __repr__(self):
        pool = self.connection_pool
        if self.host:
            host_info = ",host=%s,port=%s" % (self.host, self.port)
        else:
            host_info = ""
        s = "{}<service={}{}>".format(type(self).__name__, pool.service_name, host_info)
        return s

    async def connect_to(self, address):
        self.host, self.port = address
        await super(SentinelManagedConnection, self).connect()
        if self.connection_pool.check_connection:
            await self.send_command("PING")
            if nativestr(await self.read_response()) != "PONG":
                raise ConnectionError("PING failed")

    async def connect(self):
        if self._reader and self._writer:
            return  # already connected
        if self.connection_pool.is_master:
            await self.connect_to(await self.connection_pool.get_master_address())
        else:
            for slave in await self.connection_pool.rotate_slaves():
                try:
                    return await self.connect_to(slave)
                except ConnectionError:
                    continue
            raise SlaveNotFoundError  # Never be here

    async def read_response(self):
        try:
            return await super(SentinelManagedConnection, self).read_response()
        except ReadOnlyError:
            if self.connection_pool.is_master:
                # When talking to a master, a ReadOnlyError when likely
                # indicates that the previous master that we're still connected
                # to has been demoted to a slave and there's a new master.
                # calling disconnect will force the connection to re-query
                # sentinel during the next connect() attempt.
                self.disconnect()
                raise ConnectionError("The previous master is now a slave")
            raise


class SentinelConnectionPool(ConnectionPool):
    """
    Sentinel backed connection pool.

    If ``check_connection`` flag is set to True, SentinelManagedConnection
    sends a PING command right after establishing the connection.
    """

    def __init__(self, service_name, sentinel_manager, **kwargs):
        kwargs["connection_class"] = kwargs.get(
            "connection_class", SentinelManagedConnection
        )
        self.is_master = kwargs.pop("is_master", True)
        self.check_connection = kwargs.pop("check_connection", False)
        super(SentinelConnectionPool, self).__init__(**kwargs)
        self.connection_kwargs["connection_pool"] = weakref.proxy(self)
        self.service_name = service_name
        self.sentinel_manager = sentinel_manager

    def __repr__(self):
        return "{}<service={}({})".format(
            type(self).__name__,
            self.service_name,
            self.is_master and "master" or "slave",
        )

    def reset(self):
        super(SentinelConnectionPool, self).reset()
        self.master_address = None
        self.slave_rr_counter = None

    async def get_master_address(self):
        master_address = await self.sentinel_manager.discover_master(self.service_name)
        if self.is_master:
            if self.master_address is None:
                self.master_address = master_address
            elif master_address != self.master_address:
                # Master address changed, disconnect all clients in this pool
                self.disconnect()
        return master_address

    async def rotate_slaves(self):
        """Round-robin slave balancer"""
        slaves = await self.sentinel_manager.discover_slaves(self.service_name)
        slave_address = list()
        if slaves:
            if self.slave_rr_counter is None:
                self.slave_rr_counter = random.randint(0, len(slaves) - 1)
            for _ in range(len(slaves)):
                self.slave_rr_counter = (self.slave_rr_counter + 1) % len(slaves)
                slave_address.append(slaves[self.slave_rr_counter])
            return slave_address
        # Fallback to the master connection
        try:
            return [await self.get_master_address()]
        except MasterNotFoundError:
            pass
        raise SlaveNotFoundError("No slave found for %r" % (self.service_name))

    def _checkpid(self):
        if self.pid != os.getpid():
            self.disconnect()
            self.reset()
            self.__init__(
                self.service_name,
                self.sentinel_manager,
                is_master=self.is_master,
                check_connection=self.check_connection,
                connection_class=self.connection_class,
                max_connections=self.max_connections,
                **self.connection_kwargs
            )


class Sentinel:
    """
    Redis Sentinel cluster client

    from coredis.sentinel import Sentinel
    sentinel = Sentinel([('localhost', 26379)], stream_timeout=0.1)
    async def test():
        master = await sentinel.master_for('mymaster', stream_timeout=0.1)
        await master.set('foo', 'bar')
        slave = await sentinel.slave_for('mymaster', stream_timeout=0.1)
        await slave.get('foo')

    ``sentinels`` is a list of sentinel nodes. Each node is represented by
    a pair (hostname, port).

    ``min_other_sentinels`` defined a minimum number of peers for a sentinel.
    When querying a sentinel, if it doesn't meet this threshold, responses
    from that sentinel won't be considered valid.

    ``sentinel_kwargs`` is a dictionary of connection arguments used when
    connecting to sentinel instances. Any argument that can be passed to
    a normal Redis connection can be specified here. If ``sentinel_kwargs`` is
    not specified, any stream_timeout and socket_keepalive options specified
    in ``connection_kwargs`` will be used.

    ``connection_kwargs`` are keyword arguments that will be used when
    establishing a connection to a Redis server.
    """

    def __init__(
        self,
        sentinels,
        min_other_sentinels=0,
        sentinel_kwargs=None,
        **connection_kwargs
    ):
        # if sentinel_kwargs isn't defined, use the socket_* options from
        # connection_kwargs
        if sentinel_kwargs is None:
            sentinel_kwargs = dict(
                [
                    (k, v)
                    for k, v in iteritems(connection_kwargs)
                    if k.startswith("socket_")
                ]
            )
        self.sentinel_kwargs = sentinel_kwargs

        self.sentinels = [
            StrictRedis(hostname, port, **self.sentinel_kwargs)
            for hostname, port in sentinels
        ]
        self.min_other_sentinels = min_other_sentinels
        self.connection_kwargs = connection_kwargs

    def __repr__(self):
        sentinel_addresses = []
        for sentinel in self.sentinels:
            sentinel_addresses.append(
                "{}:{}".format(
                    sentinel.connection_pool.connection_kwargs["host"],
                    sentinel.connection_pool.connection_kwargs["port"],
                )
            )
        return "{}<sentinels=[{}]>".format(
            type(self).__name__, ",".join(sentinel_addresses)
        )

    def check_master_state(self, state, service_name):
        if not state["is_master"] or state["is_sdown"] or state["is_odown"]:
            return False
        # Check if our sentinel doesn't see other nodes
        if state["num-other-sentinels"] < self.min_other_sentinels:
            return False
        return True

    async def discover_master(self, service_name):
        """
        Asks sentinel servers for the Redis master's address corresponding
        to the service labeled ``service_name``.

        Returns a pair (address, port) or raises MasterNotFoundError if no
        master is found.
        """
        for sentinel_no, sentinel in enumerate(self.sentinels):
            try:
                masters = await sentinel.sentinel_masters()
            except (ConnectionError, TimeoutError):
                continue
            state = masters.get(service_name)
            if state and self.check_master_state(state, service_name):
                # Put this sentinel at the top of the list
                self.sentinels[0], self.sentinels[sentinel_no] = (
                    sentinel,
                    self.sentinels[0],
                )
                return state["ip"], state["port"]
        raise MasterNotFoundError("No master found for %r" % (service_name,))

    def filter_slaves(self, slaves):
        """Removes slaves that are in an ODOWN or SDOWN state"""
        slaves_alive = []
        for slave in slaves:
            if slave["is_odown"] or slave["is_sdown"]:
                continue
            slaves_alive.append((slave["ip"], slave["port"]))
        return slaves_alive

    async def discover_slaves(self, service_name):
        """Returns a list of alive slaves for service ``service_name``"""
        for sentinel in self.sentinels:
            try:
                slaves = await sentinel.sentinel_slaves(service_name)
            except (ConnectionError, ResponseError, TimeoutError):
                continue
            slaves = self.filter_slaves(slaves)
            if slaves:
                return slaves
        return []

    def master_for(
        self,
        service_name,
        redis_class=StrictRedis,
        connection_pool_class=SentinelConnectionPool,
        **kwargs
    ):
        """
        Returns a redis client instance for the ``service_name`` master.

        A SentinelConnectionPool class is used to retrive the master's
        address before establishing a new connection.

        NOTE: If the master's address has changed, any cached connections to
        the old master are closed.

        By default clients will be a redis.StrictRedis instance. Specify a
        different class to the ``redis_class`` argument if you desire
        something different.

        The ``connection_pool_class`` specifies the connection pool to use.
        The SentinelConnectionPool will be used by default.

        All other keyword arguments are merged with any connection_kwargs
        passed to this class and passed to the connection pool as keyword
        arguments to be used to initialize Redis connections.
        """
        kwargs["is_master"] = True
        connection_kwargs = dict(self.connection_kwargs)
        connection_kwargs.update(kwargs)
        return redis_class(
            connection_pool=connection_pool_class(
                service_name, self, **connection_kwargs
            )
        )

    def slave_for(
        self,
        service_name,
        redis_class=StrictRedis,
        connection_pool_class=SentinelConnectionPool,
        **kwargs
    ):
        """
        Returns redis client instance for the ``service_name`` slave(s).

        A SentinelConnectionPool class is used to retrive the slave's
        address before establishing a new connection.

        By default clients will be a redis.StrictRedis instance. Specify a
        different class to the ``redis_class`` argument if you desire
        something different.

        The ``connection_pool_class`` specifies the connection pool to use.
        The SentinelConnectionPool will be used by default.

        All other keyword arguments are merged with any connection_kwargs
        passed to this class and passed to the connection pool as keyword
        arguments to be used to initialize Redis connections.
        """
        kwargs["is_master"] = False
        connection_kwargs = dict(self.connection_kwargs)
        connection_kwargs.update(kwargs)
        return redis_class(
            connection_pool=connection_pool_class(
                service_name, self, **connection_kwargs
            )
        )



from __future__ import with_statement
import pytest
import coredis

from coredis.exceptions import ConnectionError, TimeoutError
from coredis.sentinel import (
    Sentinel,
    SentinelConnectionPool,
    MasterNotFoundError,
    SlaveNotFoundError,
)


class SentinelTestClient:
    def __init__(self, cluster, id):
        self.cluster = cluster
        self.id = id

    async def sentinel_masters(self):
        self.cluster.connection_error_if_down(self)
        self.cluster.timeout_if_down(self)
        return {self.cluster.service_name: self.cluster.master}

    async def sentinel_slaves(self, master_name):
        self.cluster.connection_error_if_down(self)
        self.cluster.timeout_if_down(self)
        if master_name != self.cluster.service_name:
            return []
        return self.cluster.slaves


class SentinelTestCluster:
    def __init__(self, service_name="mymaster", ip="127.0.0.1", port=6379):
        self.clients = {}
        self.master = {
            "ip": ip,
            "port": port,
            "is_master": True,
            "is_sdown": False,
            "is_odown": False,
            "num-other-sentinels": 0,
        }
        self.service_name = service_name
        self.slaves = []
        self.nodes_down = set()
        self.nodes_timeout = set()

    def connection_error_if_down(self, node):
        if node.id in self.nodes_down:
            raise ConnectionError

    def timeout_if_down(self, node):
        if node.id in self.nodes_timeout:
            raise TimeoutError

    def client(self, host, port, **kwargs):
        return SentinelTestClient(self, (host, port))


@pytest.fixture()
def cluster(request):
    def teardown():
        coredis.sentinel.StrictRedis = saved_StrictRedis

    cluster = SentinelTestCluster()
    saved_StrictRedis = coredis.sentinel.StrictRedis
    coredis.sentinel.StrictRedis = cluster.client
    request.addfinalizer(teardown)
    return cluster


@pytest.fixture()
def sentinel(request, cluster, event_loop):
    return Sentinel([("foo", 26379), ("bar", 26379)], loop=event_loop)


@pytest.mark.asyncio(forbid_global_loop=True)
async def test_discover_master(sentinel):
    address = await sentinel.discover_master("mymaster")
    assert address == ("127.0.0.1", 6379)


@pytest.mark.asyncio(forbid_global_loop=True)
async def test_discover_master_error(sentinel):
    with pytest.raises(MasterNotFoundError):
        await sentinel.discover_master("xxx")


@pytest.mark.asyncio(forbid_global_loop=True)
async def test_discover_master_sentinel_down(cluster, sentinel):
    # Put first sentinel 'foo' down
    cluster.nodes_down.add(("foo", 26379))
    address = await sentinel.discover_master("mymaster")
    assert address == ("127.0.0.1", 6379)
    # 'bar' is now first sentinel
    assert sentinel.sentinels[0].id == ("bar", 26379)


@pytest.mark.asyncio(forbid_global_loop=True)
async def test_discover_master_sentinel_timeout(cluster, sentinel):
    # Put first sentinel 'foo' down
    cluster.nodes_timeout.add(("foo", 26379))
    address = await sentinel.discover_master("mymaster")
    assert address == ("127.0.0.1", 6379)
    # 'bar' is now first sentinel
    assert sentinel.sentinels[0].id == ("bar", 26379)


@pytest.mark.asyncio(forbid_global_loop=True)
async def test_master_min_other_sentinels(cluster):
    sentinel = Sentinel([("foo", 26379)], min_other_sentinels=1)
    # min_other_sentinels
    with pytest.raises(MasterNotFoundError):
        await sentinel.discover_master("mymaster")
    cluster.master["num-other-sentinels"] = 2
    address = await sentinel.discover_master("mymaster")
    assert address == ("127.0.0.1", 6379)


@pytest.mark.asyncio(forbid_global_loop=True)
async def test_master_odown(cluster, sentinel):
    cluster.master["is_odown"] = True
    with pytest.raises(MasterNotFoundError):
        await sentinel.discover_master("mymaster")


@pytest.mark.asyncio(forbid_global_loop=True)
async def test_master_sdown(cluster, sentinel):
    cluster.master["is_sdown"] = True
    with pytest.raises(MasterNotFoundError):
        await sentinel.discover_master("mymaster")


@pytest.mark.asyncio(forbid_global_loop=True)
async def test_discover_slaves(cluster, sentinel):
    assert await sentinel.discover_slaves("mymaster") == []

    cluster.slaves = [
        {"ip": "slave0", "port": 1234, "is_odown": False, "is_sdown": False},
        {"ip": "slave1", "port": 1234, "is_odown": False, "is_sdown": False},
    ]
    assert await sentinel.discover_slaves("mymaster") == [
        ("slave0", 1234),
        ("slave1", 1234),
    ]

    # slave0 -> ODOWN
    cluster.slaves[0]["is_odown"] = True
    assert await sentinel.discover_slaves("mymaster") == [("slave1", 1234)]

    # slave1 -> SDOWN
    cluster.slaves[1]["is_sdown"] = True
    assert await sentinel.discover_slaves("mymaster") == []

    cluster.slaves[0]["is_odown"] = False
    cluster.slaves[1]["is_sdown"] = False

    # node0 -> DOWN
    cluster.nodes_down.add(("foo", 26379))
    assert await sentinel.discover_slaves("mymaster") == [
        ("slave0", 1234),
        ("slave1", 1234),
    ]
    cluster.nodes_down.clear()

    # node0 -> TIMEOUT
    cluster.nodes_timeout.add(("foo", 26379))
    assert await sentinel.discover_slaves("mymaster") == [
        ("slave0", 1234),
        ("slave1", 1234),
    ]


@pytest.mark.asyncio(forbid_global_loop=True)
async def test_master_for(cluster, sentinel):
    master = sentinel.master_for("mymaster")
    assert await master.ping()
    assert master.connection_pool.master_address == ("127.0.0.1", 6379)

    # Use internal connection check
    master = sentinel.master_for("mymaster", check_connection=True)
    assert await master.ping()


@pytest.mark.asyncio(forbid_global_loop=True)
async def test_slave_for(cluster, sentinel):
    cluster.slaves = [
        {"ip": "127.0.0.1", "port": 6379, "is_odown": False, "is_sdown": False},
    ]
    slave = sentinel.slave_for("mymaster")
    assert await slave.ping()


@pytest.mark.asyncio(forbid_global_loop=True)
async def test_slave_for_slave_not_found_error(cluster, sentinel):
    cluster.master["is_odown"] = True
    slave = sentinel.slave_for("mymaster", db=9)
    with pytest.raises(SlaveNotFoundError):
        await slave.ping()


@pytest.mark.asyncio(forbid_global_loop=True)
async def test_slave_round_robin(cluster, sentinel):
    cluster.slaves = [
        {"ip": "slave0", "port": 6379, "is_odown": False, "is_sdown": False},
        {"ip": "slave1", "port": 6379, "is_odown": False, "is_sdown": False},
    ]
    pool = SentinelConnectionPool("mymaster", sentinel)
    rotator = await pool.rotate_slaves()
    assert set(rotator) == {("slave0", 6379), ("slave1", 6379)}



#!/usr/bin/python
# -*- coding: utf-8 -*-

import asyncio
import pytest
import time

from coredis.cache import Cache, HerdCache


class TestCache:

    app = "test_cache"
    key = "test_key"
    data = {str(i): i for i in range(3)}

    def expensive_work(self, data):
        return data

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_set(self, r):
        await r.flushdb()
        cache = Cache(r, self.app)
        res = await cache.set(self.key, self.expensive_work(self.data), self.data)
        assert res
        identity = cache._gen_identity(self.key, self.data)
        content = await r.get(identity)
        content = cache._unpack(content)
        assert content == self.data

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_set_timeout(self, r, event_loop):
        await r.flushdb()
        cache = Cache(r, self.app)
        res = await cache.set(
            self.key, self.expensive_work(self.data), self.data, expire_time=1
        )
        assert res
        identity = cache._gen_identity(self.key, self.data)
        content = await r.get(identity)
        content = cache._unpack(content)
        assert content == self.data
        await asyncio.sleep(1.1)
        content = await r.get(identity)
        assert content is None

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_set_with_plain_key(self, r):
        await r.flushdb()
        cache = Cache(r, self.app, identity_generator_class=None)
        res = await cache.set(
            self.key, self.expensive_work(self.data), self.data, expire_time=1
        )
        assert res
        identity = cache._gen_identity(self.key, self.data)
        assert identity == self.key
        content = await r.get(identity)
        content = cache._unpack(content)
        assert content == self.data

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_get(self, r):
        await r.flushdb()
        cache = Cache(r, self.app)
        res = await cache.set(
            self.key, self.expensive_work(self.data), self.data, expire_time=1
        )
        assert res
        content = await cache.get(self.key, self.data)
        assert content == self.data

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_set_many(self, r):
        await r.flushdb()
        cache = Cache(r, self.app)
        res = await cache.set_many(self.expensive_work(self.data), self.data)
        assert res
        for key, value in self.data.items():
            assert await cache.get(key, self.data) == value

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_delete(self, r):
        await r.flushdb()
        cache = Cache(r, self.app)
        res = await cache.set(
            self.key, self.expensive_work(self.data), self.data, expire_time=1
        )
        assert res
        content = await cache.get(self.key, self.data)
        assert content == self.data
        res = await cache.delete(self.key, self.data)
        assert res
        content = await cache.get(self.key, self.data)
        assert content is None

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_delete_pattern(self, r):
        await r.flushdb()
        cache = Cache(r, self.app)
        await cache.set_many(self.expensive_work(self.data), self.data)
        res = await cache.delete_pattern("test_*", 10)
        assert res == 3
        content = await cache.get(self.key, self.data)
        assert content is None

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_ttl(self, r, event_loop):
        await r.flushdb()
        cache = Cache(r, self.app)
        await cache.set(
            self.key, self.expensive_work(self.data), self.data, expire_time=1
        )
        ttl = await cache.ttl(self.key, self.data)
        assert ttl > 0
        await asyncio.sleep(1.1)
        ttl = await cache.ttl(self.key, self.data)
        assert ttl < 0

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_exists(self, r, event_loop):
        await r.flushdb()
        cache = Cache(r, self.app)
        await cache.set(
            self.key, self.expensive_work(self.data), self.data, expire_time=1
        )
        exists = await cache.exist(self.key, self.data)
        assert exists is True
        await asyncio.sleep(1.1)
        exists = await cache.exist(self.key, self.data)
        assert exists is False


class TestHerdCache:

    app = "test_cache"
    key = "test_key"
    data = {str(i): i for i in range(3)}

    def expensive_work(self, data):
        return data

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_set(self, r):
        await r.flushdb()
        cache = HerdCache(r, self.app, default_herd_timeout=1, extend_herd_timeout=1)
        now = int(time.time())
        res = await cache.set(self.key, self.expensive_work(self.data), self.data)
        assert res
        identity = cache._gen_identity(self.key, self.data)
        content = await r.get(identity)
        content, expect_expire_time = cache._unpack(content)
        # supposed equal to 1, but may there be latency
        assert expect_expire_time - now <= 1
        assert content == self.data

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_get(self, r):
        await r.flushdb()
        cache = HerdCache(r, self.app, default_herd_timeout=1, extend_herd_timeout=1)
        res = await cache.set(self.key, self.expensive_work(self.data), self.data)
        assert res
        content = await cache.get(self.key, self.data)
        assert content == self.data

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_set_many(self, r):
        await r.flushdb()
        cache = HerdCache(r, self.app, default_herd_timeout=1, extend_herd_timeout=1)
        res = await cache.set_many(self.expensive_work(self.data), self.data)
        assert res
        for key, value in self.data.items():
            assert await cache.get(key, self.data) == value

    @pytest.mark.asyncio(forbid_global_loop=True)
    async def test_herd(self, r, event_loop):
        await r.flushdb()
        now = int(time.time())
        cache = HerdCache(r, self.app, default_herd_timeout=1, extend_herd_timeout=1)
        await cache.set(self.key, self.expensive_work(self.data), self.data)
        await asyncio.sleep(1)
        # first get
        identity = cache._gen_identity(self.key, self.data)
        content = await r.get(identity)
        content, expect_expire_time = cache._unpack(content)
        assert now + 1 == expect_expire_time
        # HerdCach.get
        await asyncio.sleep(0.1)
        res = await cache.get(self.key, self.data)
        # first herd get will reset expire time and return None
        assert res is None
        # second get
        identity = cache._gen_identity(self.key, self.data)
        content = await r.get(identity)
        content, new_expire_time = cache._unpack(content)
        assert new_expire_time >= expect_expire_time + 1



class RedisError(Exception):
    pass


class AuthenticationError(RedisError):
    pass


class ConnectionError(RedisError):
    pass


class TimeoutError(RedisError):
    pass


class BusyLoadingError(ConnectionError):
    pass


class InvalidResponse(RedisError):
    pass


class ResponseError(RedisError):
    pass


class DataError(RedisError):
    pass


class PubSubError(RedisError):
    pass


class WatchError(RedisError):
    pass


class NoScriptError(ResponseError):
    pass


class ExecAbortError(ResponseError):
    pass


class ReadOnlyError(ResponseError):
    pass


class LockError(RedisError, ValueError):
    """Errors acquiring or releasing a lock"""

    # NOTE: For backwards compatability, this class derives from ValueError.
    # This was originally chosen to behave like threading.Lock.
    pass


class CacheError(RedisError):
    """Basic error of coredis.cache"""

    pass


class SerializeError(CacheError):
    pass


class CompressError(CacheError):
    pass


class RedisClusterException(Exception):
    pass


class RedisClusterError(Exception):
    pass


class ClusterDownException(Exception):
    pass


class ClusterError(RedisError):
    pass


class ClusterCrossSlotError(ResponseError):
    message = "Keys in request don't hash to the same slot"


class ClusterDownError(ClusterError, ResponseError):
    def __init__(self, resp):
        self.args = (resp,)
        self.message = resp


class ClusterTransactionError(ClusterError):
    def __init__(self, msg):
        self.msg = msg


class AskError(ResponseError):
    """
    src node: MIGRATING to dst node
        get > ASK error
        ask dst node > ASKING command
    dst node: IMPORTING from src node
        asking command only affects next command
        any op will be allowed after asking command
    """

    def __init__(self, resp):
        """should only redirect to master node"""
        self.args = (resp,)
        self.message = resp
        slot_id, new_node = resp.split(" ")
        host, port = new_node.rsplit(":", 1)
        self.slot_id = int(slot_id)
        self.node_addr = self.host, self.port = host, int(port)


class TryAgainError(ResponseError):
    def __init__(self, *args, **kwargs):
        pass


class MovedError(AskError):
    pass



from coredis.pubsub import PubSub, ClusterPubSub
from coredis.utils import dict_merge, merge_result, list_keys_to_dict, NodeFlag


def parse_pubsub_numsub(response, **options):
    return list(zip(response[0::2], response[1::2]))


class PubSubCommandMixin:

    RESPONSE_CALLBACKS = {
        "PUBSUB NUMSUB": parse_pubsub_numsub,
    }

    def pubsub(self, **kwargs):
        """
        Return a Publish/Subscribe object. With this object, you can
        subscribe to channels and listen for messages that get published to
        them.
        """
        return PubSub(self.connection_pool, **kwargs)

    async def publish(self, channel, message):
        """
        Publish ``message`` on ``channel``.
        Returns the number of subscribers the message was delivered to.
        """
        return await self.execute_command("PUBLISH", channel, message)

    async def pubsub_channels(self, pattern="*"):
        """
        Return a list of channels that have at least one subscriber
        """
        return await self.execute_command("PUBSUB CHANNELS", pattern)

    async def pubsub_numpat(self):
        """
        Returns the number of subscriptions to patterns
        """
        return await self.execute_command("PUBSUB NUMPAT")

    async def pubsub_numsub(self, *args):
        """
        Return a list of (channel, number of subscribers) tuples
        for each channel given in ``*args``
        """
        return await self.execute_command("PUBSUB NUMSUB", *args)


def parse_cluster_pubsub_channels(res, **options):
    """
    Result callback, handles different return types
    switchable by the `aggregate` flag.
    """
    aggregate = options.get("aggregate", True)
    if not aggregate:
        return res
    return merge_result(res)


def parse_cluster_pubsub_numpat(res, **options):
    """
    Result callback, handles different return types
    switchable by the `aggregate` flag.
    """
    aggregate = options.get("aggregate", True)
    if not aggregate:
        return res

    numpat = 0
    for node, node_numpat in res.items():
        numpat += node_numpat
    return numpat


def parse_cluster_pubsub_numsub(res, **options):
    """
    Result callback, handles different return types
    switchable by the `aggregate` flag.
    """
    aggregate = options.get("aggregate", True)
    if not aggregate:
        return res

    numsub_d = dict()
    for _, numsub_tups in res.items():
        for channel, numsubbed in numsub_tups:
            try:
                numsub_d[channel] += numsubbed
            except KeyError:
                numsub_d[channel] = numsubbed

    ret_numsub = []
    for channel, numsub in numsub_d.items():
        ret_numsub.append((channel, numsub))
    return ret_numsub


class CLusterPubSubCommandMixin(PubSubCommandMixin):

    NODES_FLAGS = dict_merge(
        list_keys_to_dict(
            ["PUBSUB CHANNELS", "PUBSUB NUMSUB", "PUBSUB NUMPAT"], NodeFlag.ALL_NODES
        )
    )

    RESULT_CALLBACKS = dict_merge(
        list_keys_to_dict(["PUBSUB CHANNELS",], parse_cluster_pubsub_channels),
        list_keys_to_dict(["PUBSUB NUMSUB",], parse_cluster_pubsub_numsub),
        list_keys_to_dict(["PUBSUB NUMPAT",], parse_cluster_pubsub_numpat),
    )

    def pubsub(self, **kwargs):
        return ClusterPubSub(self.connection_pool, **kwargs)



# -*- coding: utf-8 -*-

# python std lib
from __future__ import with_statement

# rediscluster imports
import asyncio
from coredis.exceptions import RedisClusterException, NoScriptError, ResponseError
from coredis.utils import b

# 3rd party imports
import pytest


multiply_script = """
local value = redis.call('GET', KEYS[1])
value = tonumber(value)
return value * ARGV[1]"""

msgpack_hello_script = """
local message = cmsgpack.unpack(ARGV[1])
local name = message['name']
return "hello " .. name
"""
msgpack_hello_script_broken = """
local message = cmsgpack.unpack(ARGV[1])
local names = message['name']
return "hello " .. name
"""


class TestScripting:
    async def reset_scripts(self, r):
        await r.script_flush()

    @pytest.mark.asyncio()
    async def test_eval(self, r):
        await r.set("a", 2)
        # 2 * 3 == 6
        assert await r.eval(multiply_script, 1, "a", 3) == 6

    @pytest.mark.asyncio()
    async def test_eval_same_slot(self, r):
        await r.set("A{foo}", 2)
        await r.set("B{foo}", 4)
        # 2 * 4 == 8

        script = """
        local value = redis.call('GET', KEYS[1])
        local value2 = redis.call('GET', KEYS[2])
        return value * value2
        """
        result = await r.eval(script, 2, "A{foo}", "B{foo}")
        assert result == 8

    @pytest.mark.asyncio()
    async def test_eval_crossslot(self, r):
        """
        This test assumes that {foo} and {bar} will not go to the same
        server when used. In 3 masters + 3 slaves config this should pass.
        """
        await r.set("A{foo}", 2)
        await r.set("B{bar}", 4)
        # 2 * 4 == 8

        script = """
        local value = redis.call('GET', KEYS[1])
        local value2 = redis.call('GET', KEYS[2])
        return value * value2
        """
        with pytest.raises(RedisClusterException):
            await r.eval(script, 2, "A{foo}", "B{bar}")

    @pytest.mark.asyncio()
    async def test_evalsha(self, r):
        await r.set("a", 2)
        sha = await r.script_load(multiply_script)
        # 2 * 3 == 6
        assert await r.evalsha(sha, 1, "a", 3) == 6

    @pytest.mark.asyncio()
    async def test_evalsha_script_not_loaded(self, r):
        await r.set("a", 2)
        sha = await r.script_load(multiply_script)
        # remove the script from Redis's cache
        await r.script_flush()
        with pytest.raises(NoScriptError):
            await r.evalsha(sha, 1, "a", 3)

    @pytest.mark.asyncio()
    async def test_script_loading(self, r):
        # get the sha, then clear the cache
        sha = await r.script_load(multiply_script)
        await r.script_flush()
        assert await r.script_exists(sha) == [False]
        await r.script_load(multiply_script)
        assert await r.script_exists(sha) == [True]

    @pytest.mark.asyncio()
    async def test_script_object(self, r):
        await r.set("a", 2)
        multiply = r.register_script(multiply_script)
        assert multiply.sha == "29cdf3e36c89fa05d7e6d6b9734b342ab15c9ea7"
        # test evalsha fail -> script load + retry
        assert await multiply.execute(keys=["a"], args=[3]) == 6
        assert multiply.sha
        assert await r.script_exists(multiply.sha) == [True]
        # test first evalsha
        assert await multiply.execute(keys=["a"], args=[3]) == 6

    @pytest.mark.asyncio(forbid_global_loop=True)
    @pytest.mark.xfail(reason="Not Yet Implemented")
    async def test_script_object_in_pipeline(self, r):
        multiply = await r.register_script(multiply_script)
        assert not multiply.sha
        pipe = r.pipeline()
        await pipe.set("a", 2)
        await pipe.get("a")
        multiply(keys=["a"], args=[3], client=pipe)
        # even though the pipeline wasn't executed yet, we made sure the
        # script was loaded and got a valid sha
        assert multiply.sha
        assert await r.script_exists(multiply.sha) == [True]
        # [SET worked, GET 'a', result of multiple script]
        assert await pipe.execute() == [True, b("2"), 6]

        # purge the script from redis's cache and re-run the pipeline
        # the multiply script object knows it's sha, so it shouldn't get
        # reloaded until pipe.execute()
        await r.script_flush()
        pipe = await r.pipeline()
        await pipe.set("a", 2)
        await pipe.get("a")
        assert multiply.sha
        multiply(keys=["a"], args=[3], client=pipe)
        assert await r.script_exists(multiply.sha) == [False]
        # [SET worked, GET 'a', result of multiple script]
        assert await pipe.execute() == [True, b("2"), 6]

    @pytest.mark.asyncio(forbid_global_loop=True)
    @pytest.mark.xfail(reason="Not Yet Implemented")
    async def test_eval_msgpack_pipeline_error_in_lua(self, r):
        msgpack_hello = await r.register_script(msgpack_hello_script)
        assert not msgpack_hello.sha

        pipe = r.pipeline()

        # avoiding a dependency to msgpack, this is the output of
        # msgpack.dumps({"name": "joe"})
        msgpack_message_1 = b"\x81\xa4name\xa3Joe"

        msgpack_hello(args=[msgpack_message_1], client=pipe)

        assert await r.script_exists(msgpack_hello.sha) == [True]
        assert await pipe.execute()[0] == b"hello Joe"

        msgpack_hello_broken = await r.register_script(msgpack_hello_script_broken)

        msgpack_hello_broken(args=[msgpack_message_1], client=pipe)
        with pytest.raises(ResponseError) as excinfo:
            await pipe.execute()
        assert excinfo.type == ResponseError



from coredis.utils import NodeFlag, bool_ok, nativestr


class ConnectionCommandMixin:

    RESPONSE_CALLBACKS = {
        "AUTH": bool,
        "PING": lambda r: nativestr(r) == "PONG",
        "SELECT": bool_ok,
    }

    async def echo(self, value):
        "Echo the string back from the server"
        return await self.execute_command("ECHO", value)

    async def ping(self):
        "Ping the Redis server"
        return await self.execute_command("PING")


class ClusterConnectionCommandMixin(ConnectionCommandMixin):

    NODES_FLAGS = {"PING": NodeFlag.ALL_NODES, "ECHO": NodeFlag.ALL_NODES}

    RESULT_CALLBACKS = {"ECHO": lambda res: res, "PING": lambda res: res}



# -*- coding: utf-8 -*-

# python std lib
from __future__ import with_statement

# rediscluster imports
from coredis.commands.cluster import parse_cluster_slots
from coredis.exceptions import RedisClusterException, ClusterDownError
from coredis.utils import (
    list_keys_to_dict,
    b,
    dict_merge,
    blocked_command,
    merge_result,
    first_key,
    clusterdown_wrapper,
)

# 3rd party imports
import pytest


def test_parse_cluster_slots():
    """
    Example raw output from redis cluster. Output is form a redis 3.2.x node
    that includes the id in the reponse. The test below that do not include the id
    is to validate that the code is compatible with redis versions that do not contain
    that value in the response from the server.

    127.0.0.1:10000> cluster slots
    1) 1) (integer) 5461
       2) (integer) 10922
       3) 1) "10.0.0.1"
          2) (integer) 10000
          3) "3588b4cf9fc72d57bb262a024747797ead0cf7ea"
       4) 1) "10.0.0.4"
          2) (integer) 10000
          3) "a72c02c7d85f4ec3145ab2c411eefc0812aa96b0"
    2) 1) (integer) 10923
       2) (integer) 16383
       3) 1) "10.0.0.2"
          2) (integer) 10000
          3) "ffd36d8d7cb10d813f81f9662a835f6beea72677"
       4) 1) "10.0.0.5"
          2) (integer) 10000
          3) "5c15b69186017ddc25ebfac81e74694fc0c1a160"
    3) 1) (integer) 0
       2) (integer) 5460
       3) 1) "10.0.0.3"
          2) (integer) 10000
          3) "069cda388c7c41c62abe892d9e0a2d55fbf5ffd5"
       4) 1) "10.0.0.6"
          2) (integer) 10000
          3) "dc152a08b4cf1f2a0baf775fb86ad0938cb907dc"
    """

    extended_mock_response = [
        [
            0,
            5460,
            ["172.17.0.2", 7000, "ffd36d8d7cb10d813f81f9662a835f6beea72677"],
            ["172.17.0.2", 7003, "5c15b69186017ddc25ebfac81e74694fc0c1a160"],
        ],
        [
            5461,
            10922,
            ["172.17.0.2", 7001, "069cda388c7c41c62abe892d9e0a2d55fbf5ffd5"],
            ["172.17.0.2", 7004, "dc152a08b4cf1f2a0baf775fb86ad0938cb907dc"],
        ],
        [
            10923,
            16383,
            ["172.17.0.2", 7002, "3588b4cf9fc72d57bb262a024747797ead0cf7ea"],
            ["172.17.0.2", 7005, "a72c02c7d85f4ec3145ab2c411eefc0812aa96b0"],
        ],
    ]

    parse_cluster_slots(extended_mock_response)

    extended_mock_binary_response = [
        [
            0,
            5460,
            [b("172.17.0.2"), 7000, b("ffd36d8d7cb10d813f81f9662a835f6beea72677")],
            [b("172.17.0.2"), 7003, b("5c15b69186017ddc25ebfac81e74694fc0c1a160")],
        ],
        [
            5461,
            10922,
            [b("172.17.0.2"), 7001, b("069cda388c7c41c62abe892d9e0a2d55fbf5ffd5")],
            [b("172.17.0.2"), 7004, b("dc152a08b4cf1f2a0baf775fb86ad0938cb907dc")],
        ],
        [
            10923,
            16383,
            [b("172.17.0.2"), 7002, b("3588b4cf9fc72d57bb262a024747797ead0cf7ea")],
            [b("172.17.0.2"), 7005, b("a72c02c7d85f4ec3145ab2c411eefc0812aa96b0")],
        ],
    ]

    extended_mock_parsed = {
        (0, 5460): [
            {
                "host": b"172.17.0.2",
                "node_id": b"ffd36d8d7cb10d813f81f9662a835f6beea72677",
                "port": 7000,
                "server_type": "master",
            },
            {
                "host": b"172.17.0.2",
                "node_id": b"5c15b69186017ddc25ebfac81e74694fc0c1a160",
                "port": 7003,
                "server_type": "slave",
            },
        ],
        (5461, 10922): [
            {
                "host": b"172.17.0.2",
                "node_id": b"069cda388c7c41c62abe892d9e0a2d55fbf5ffd5",
                "port": 7001,
                "server_type": "master",
            },
            {
                "host": b"172.17.0.2",
                "node_id": b"dc152a08b4cf1f2a0baf775fb86ad0938cb907dc",
                "port": 7004,
                "server_type": "slave",
            },
        ],
        (10923, 16383): [
            {
                "host": b"172.17.0.2",
                "node_id": b"3588b4cf9fc72d57bb262a024747797ead0cf7ea",
                "port": 7002,
                "server_type": "master",
            },
            {
                "host": b"172.17.0.2",
                "node_id": b"a72c02c7d85f4ec3145ab2c411eefc0812aa96b0",
                "port": 7005,
                "server_type": "slave",
            },
        ],
    }

    assert parse_cluster_slots(extended_mock_binary_response) == extended_mock_parsed


def test_list_keys_to_dict():
    def mock_true():
        return True

    assert list_keys_to_dict(["FOO", "BAR"], mock_true) == {
        "FOO": mock_true,
        "BAR": mock_true,
    }


def test_dict_merge():
    a = {"a": 1}
    b = {"b": 2}
    c = {"c": 3}
    assert dict_merge(a, b, c) == {"a": 1, "b": 2, "c": 3}


def test_dict_merge_empty_list():
    assert dict_merge([]) == {}


def test_blocked_command():
    with pytest.raises(RedisClusterException) as ex:
        blocked_command(None, "SET")
    assert str(ex.value) == "Command: SET is blocked in redis cluster mode"


def test_merge_result():
    assert merge_result({"a": [1, 2, 3], "b": [4, 5, 6]}) == [1, 2, 3, 4, 5, 6]
    assert merge_result({"a": [1, 2, 3], "b": [1, 2, 3]}) == [1, 2, 3]


def test_merge_result_value_error():
    with pytest.raises(ValueError):
        merge_result([])


def test_first_key():
    assert first_key({"foo": 1}) == 1

    with pytest.raises(RedisClusterException) as ex:
        first_key({"foo": 1, "bar": 2})
    assert str(ex.value).startswith("More then 1 result from command")


def test_first_key_value_error():
    with pytest.raises(ValueError):
        first_key(None)


@pytest.mark.asyncio(forbid_global_loop=True)
async def test_clusterdown_wrapper():
    @clusterdown_wrapper
    def bad_func():
        raise ClusterDownError("CLUSTERDOWN")

    with pytest.raises(ClusterDownError) as cex:
        await bad_func()
    assert str(cex.value).startswith("CLUSTERDOWN error. Unable to rebuild the cluster")



from coredis.utils import bool_ok, nativestr, NodeFlag, list_keys_to_dict, dict_merge
from coredis.exceptions import RedisError, ClusterError


def parse_cluster_info(response, **options):
    response = nativestr(response)
    return dict([line.split(":") for line in response.splitlines() if line])


def parse_cluster_nodes(resp, **options):
    """
    @see: http://redis.io/commands/cluster-nodes  # string
    @see: http://redis.io/commands/cluster-slaves # list of string
    """
    if isinstance(resp, list):
        resp = [nativestr(row) for row in resp]
    else:
        resp = nativestr(resp)
    current_host = options.get("current_host", "")

    def parse_slots(s):
        slots, migrations = [], []
        for r in s.split(" "):
            if "->-" in r:
                slot_id, dst_node_id = r[1:-1].split("->-", 1)
                migrations.append(
                    {"slot": int(slot_id), "node_id": dst_node_id, "state": "migrating"}
                )
            elif "-<-" in r:
                slot_id, src_node_id = r[1:-1].split("-<-", 1)
                migrations.append(
                    {"slot": int(slot_id), "node_id": src_node_id, "state": "importing"}
                )
            elif "-" in r:
                start, end = r.split("-")
                slots.extend(range(int(start), int(end) + 1))
            else:
                slots.append(int(r))

        return slots, migrations

    if isinstance(resp, str):
        resp = resp.splitlines()

    nodes = []
    for line in resp:
        parts = line.split(" ", 8)
        (
            self_id,
            addr,
            flags,
            master_id,
            ping_sent,
            pong_recv,
            config_epoch,
            link_state,
        ) = parts[:8]

        host, port = addr.rsplit(":", 1)

        node = {
            "id": self_id,
            "host": host or current_host,
            "port": int(port.split("@")[0]),
            "flags": tuple(flags.split(",")),
            "master": master_id if master_id != "-" else None,
            "ping-sent": int(ping_sent),
            "pong-recv": int(pong_recv),
            "link-state": link_state,
            "slots": [],
            "migrations": [],
        }

        if len(parts) >= 9:
            slots, migrations = parse_slots(parts[8])
            node["slots"], node["migrations"] = tuple(slots), migrations

        nodes.append(node)

    return nodes


def parse_cluster_slots(response):
    res = dict()
    for slot_info in response:
        min_slot, max_slot = slot_info[:2]
        nodes = slot_info[2:]
        parse_node = lambda node: {
            "host": node[0],
            "port": node[1],
            "node_id": node[2] if len(node) > 2 else "",
            "server_type": "slave",
        }
        res[(min_slot, max_slot)] = [parse_node(node) for node in nodes]
        res[(min_slot, max_slot)][0]["server_type"] = "master"
    return res


class ClusterCommandMixin:

    NODES_FLAGS = dict_merge(
        {
            "CLUSTER INFO": NodeFlag.ALL_NODES,
            "CLUSTER COUNTKEYSINSLOT": NodeFlag.SLOT_ID,
        },
        list_keys_to_dict(["CLUSTER NODES", "CLUSTER SLOTS"], NodeFlag.RANDOM),
    )

    RESPONSE_CALLBACKS = {
        "CLUSTER ADDSLOTS": bool_ok,
        "CLUSTER COUNT-FAILURE-REPORTS": lambda x: int(x),
        "CLUSTER COUNTKEYSINSLOT": lambda x: int(x),
        "CLUSTER DELSLOTS": bool_ok,
        "CLUSTER FAILOVER": bool_ok,
        "CLUSTER FORGET": bool_ok,
        "CLUSTER INFO": parse_cluster_info,
        "CLUSTER KEYSLOT": lambda x: int(x),
        "CLUSTER MEET": bool_ok,
        "CLUSTER NODES": parse_cluster_nodes,
        "CLUSTER REPLICATE": bool_ok,
        "CLUSTER RESET": bool_ok,
        "CLUSTER SAVECONFIG": bool_ok,
        "CLUSTER SET-CONFIG-EPOCH": bool_ok,
        "CLUSTER SETSLOT": bool_ok,
        "CLUSTER SLAVES": parse_cluster_nodes,
        "CLUSTER SLOTS": parse_cluster_slots,
        "ASKING": bool_ok,
        "READONLY": bool_ok,
        "READWRITE": bool_ok,
    }

    RESULT_CALLBACKS = dict_merge(
        list_keys_to_dict(
            [
                "CLUSTER INFO",
                "CLUSTER ADDSLOTS",
                "CLUSTER COUNT-FAILURE-REPORTS",
                "CLUSTER DELSLOTS",
                "CLUSTER FAILOVER",
                "CLUSTER FORGET",
            ],
            lambda res: res,
        )
    )

    def _nodes_slots_to_slots_nodes(self, mapping):
        """
        Converts a mapping of
        {id: <node>, slots: (slot1, slot2)}
        to
        {slot1: <node>, slot2: <node>}

        Operation is expensive so use with caution
        """
        out = {}
        for node in mapping:
            for slot in node["slots"]:
                out[str(slot)] = node["id"]
        return out

    async def cluster_addslots(self, node_id, *slots):
        """
        Assign new hash slots to receiving node

        Sends to specefied node
        """
        return await self.execute_command("CLUSTER ADDSLOTS", *slots, node_id=node_id)

    async def cluster_count_failure_report(self, node_id=""):
        """
        Return the number of failure reports active for a given node

        Sends to specefied node
        """
        return await self.execute_command(
            "CLUSTER COUNT-FAILURE-REPORTS", node_id=node_id
        )

    async def cluster_countkeysinslot(self, slot_id):
        """
        Return the number of local keys in the specified hash slot

        Send to node based on specefied slot_id
        """
        return await self.execute_command("CLUSTER COUNTKEYSINSLOT", slot_id)

    async def cluster_delslots(self, *slots):
        """
        Set hash slots as unbound in the cluster.
        It determines by it self what node the slot is in and sends it there

        Returns a list of the results for each processed slot.
        """
        cluster_nodes = self._nodes_slots_to_slots_nodes(await self.cluster_nodes())
        res = list()
        for slot in slots:
            res.append(
                await self.execute_command(
                    "CLUSTER DELSLOTS", slot, node_id=cluster_nodes[slot]
                )
            )
        return res

    async def cluster_failover(self, node_id, option):
        """
        Forces a slave to perform a manual failover of its master

        Sends to specefied node
        """
        if not isinstance(option, str) or option.upper() not in {"FORCE", "TAKEOVER"}:
            raise ClusterError("Wrong option provided")
        return await self.execute_command("CLUSTER FAILOVER", option, node_id=node_id)

    async def cluster_forget(self, node_id):
        """
        remove a node via its node ID from the set of known nodes
        of the Redis Cluster node receiving the command

        Sends to all nodes in the cluster
        """
        return await self.execute_command("CLUSTER FORGET", node_id)

    async def cluster_info(self):
        """
        Provides info about Redis Cluster node state

        Sends to random node in the cluster
        """
        return await self.execute_command("CLUSTER INFO")

    async def cluster_keyslot(self, name):
        """
        Returns the hash slot of the specified key

        Sends to random node in the cluster
        """
        return await self.execute_command("CLUSTER KEYSLOT", name)

    async def cluster_meet(self, node_id, host, port):
        """
        Force a node cluster to handshake with another node.

        Sends to specefied node
        """
        return await self.execute_command("CLUSTER MEET", host, port, node_id=node_id)

    async def cluster_nodes(self):
        """
        Force a node cluster to handshake with another node

        Sends to random node in the cluster
        """
        return await self.execute_command("CLUSTER NODES")

    async def cluster_replicate(self, target_node_id):
        """
        Reconfigure a node as a slave of the specified master node

        Sends to specefied node
        """
        return await self.execute_command("CLUSTER REPLICATE", target_node_id)

    async def cluster_reset(self, node_id, soft=True):
        """
        Reset a Redis Cluster node

        If 'soft' is True then it will send 'SOFT' argument
        If 'soft' is False then it will send 'HARD' argument

        Sends to specefied node
        """
        option = "SOFT" if soft else "HARD"
        return await self.execute_command("CLUSTER RESET", option, node_id=node_id)

    async def cluster_reset_all_nodes(self, soft=True):
        """
        Send CLUSTER RESET to all nodes in the cluster

        If 'soft' is True then it will send 'SOFT' argument
        If 'soft' is False then it will send 'HARD' argument

        Sends to all nodes in the cluster
        """
        option = "SOFT" if soft else "HARD"
        res = list()
        for node in await self.cluster_nodes():
            res.append(
                await self.execute_command("CLUSTER RESET", option, node_id=node["id"])
            )
        return res

    async def cluster_save_config(self):
        """
        Forces the node to save cluster state on disk

        Sends to all nodes in the cluster
        """
        return await self.execute_command("CLUSTER SAVECONFIG")

    async def cluster_set_config_epoch(self, node_id, epoch):
        """
        Set the configuration epoch in a new node

        Sends to specefied node
        """
        return await self.execute_command(
            "CLUSTER SET-CONFIG-EPOCH", epoch, node_id=node_id
        )

    async def cluster_setslot(self, node_id, slot_id, state):
        """
        Bind an hash slot to a specific node

        Sends to specified node
        """
        if state.upper() in {"IMPORTING", "MIGRATING", "NODE"} and node_id is not None:
            return await self.execute_command(
                "CLUSTER SETSLOT", slot_id, state, node_id
            )
        elif state.upper() == "STABLE":
            return await self.execute_command("CLUSTER SETSLOT", slot_id, "STABLE")
        else:
            raise RedisError("Invalid slot state: {0}".format(state))

    async def cluster_get_keys_in_slot(self, slot_id, count):
        """
        Return local key names in the specified hash slot
        Sends to specified node
        """
        return await self.execute_command("CLUSTER GETKEYSINSLOT", slot_id, count)

    async def cluster_slaves(self, target_node_id):
        """
        Force a node cluster to handshake with another node

        Sends to targeted cluster node
        """
        return await self.execute_command("CLUSTER SLAVES", target_node_id)

    async def cluster_slots(self):
        """
        Get array of Cluster slot to node mappings

        Sends to random node in the cluster
        """
        return await self.execute_command("CLUSTER SLOTS")
